//  $Id: prom_list.hh,v 1.4 2004/08/04 15:57:15 adams Exp $
//  Author:             Mark F. Adams
// Copyright (c) 2000 by Mark F. Adams 
//  Filename:           prom_list.hh
//
#ifndef __LIST_H__
#define __LIST_H__

typedef unsigned char       BYTE;
typedef unsigned int        UINT;

struct CPlex // warning variable length structure
{
public:
  CPlex* pNext;
  UINT   nMax;
  UINT   nCur;
  /* BYTE data[maxNum*elementSize]; */
  
  void* data() { return this+1; }
  
  static CPlex* create(CPlex*& head, UINT nMax, UINT cbElement);
  void freeDataChain();       // free this one and links
};

typedef void* Position;  // abstract iteration position

template <class T> struct CNode
{
  CNode<T>* pNext;
  CNode<T>* pPrev;
  T data;
};

/////////////////////////////////////////////////////////////////////////////
// PromList
//
template <class T> class PromList 
{
public:
  // Construction
  PromList( int nBlockSize = 10 ){
    assert(this!=NULL); assert(nBlockSize > 0);	  
    nCount_ = 0;
    pNodeHead_ = pNodeTail_ = pNodeFree_ = NULL;
    pBlocks_ = NULL;
    nBlockSize_ = nBlockSize;  assert(nBlockSize_ > 0);
  }
  //
  virtual ~PromList(){
    assert(this!=NULL);
    removeAll();
  }
  // count of elements
  int getCount() const {assert(this!=NULL); return nCount_; }
  bool isEmpty() const {assert(this!=NULL); return (nCount_ == 0); };  
  // peek at head or tail
  T getHead() const{
    assert(this!=NULL);
    assert(pNodeHead_ != NULL); 
    return pNodeHead_->data;
  }
  T getTail() const{
    assert(this!=NULL);
    assert(pNodeTail_ != NULL);
    return pNodeTail_->data;
  }
  // get head or tail (and remove it) - don't call on empty list!
  T removeHead() {
    assert(this != NULL);
    assert(pNodeHead_ != NULL);  // don't call on empty list !!!
    
    CNode<T>* pOldNode = pNodeHead_;
    T returnValue = pOldNode->data; //assert(returnValue != 0);
    
    pNodeHead_ = pOldNode->pNext;
    if (pNodeHead_ != NULL)
      pNodeHead_->pPrev = NULL;
    else
      pNodeTail_ = NULL;
    freeNode(pOldNode);
    return returnValue;
  }
  // 
  T removeTail() {
    assert(this != NULL);
    assert(pNodeTail_ != NULL);  // don't call on empty list !!!

    CNode<T>* pOldNode = pNodeTail_;
    T returnValue = pOldNode->data; //assert(returnValue != 0);
    
    pNodeTail_ = pOldNode->pPrev;
    if (pNodeTail_ != NULL)
      pNodeTail_->pNext = NULL;
    else
      pNodeHead_ = NULL;
    freeNode(pOldNode);
    return returnValue;
  }
  // add before head or after tail
  Position addHead(T newElement){
    assert(this != NULL); //assert(newElement != 0);
    PetscFunctionBegin;
    CNode<T>* pNewNode = newNode(NULL, pNodeHead_);
    pNewNode->data = newElement;
    if (pNodeHead_ != NULL)    pNodeHead_->pPrev = pNewNode;
    else                        pNodeTail_ = pNewNode;
    pNodeHead_ = pNewNode;
    PetscFunctionReturn((Position)pNewNode);
  }
  // 
  Position addTail(T newElement){
    assert(this != NULL); //assert(newElement != 0);
    PetscFunctionBegin;
    CNode<T>* pNewNode = newNode(pNodeTail_, NULL);
    pNewNode->data = newElement;
    if (pNodeTail_ != NULL)
      pNodeTail_->pNext = pNewNode;
    else
      pNodeHead_ = pNewNode;
    pNodeTail_ = pNewNode;
    PetscFunctionReturn((Position)pNewNode);
  }
  // add another list of elements before head or after tail
  void addHead(PromList<T>* pNewList){
    assert(this != NULL);  assert(pNewList != NULL);
    // add a list of same elements to head (maintain order)
    Position pos = pNewList->getTailPosition();
    while (pos != NULL)
      addHead(pNewList->getPrev(pos));
  }
  //
  void addTail(PromList<T>* pNewList){
    assert(this != NULL);  assert(pNewList != NULL);    
    // add a list of same elements
    Position pos = pNewList->getHeadPosition();
    while (pos != NULL) addTail( pNewList->getNext(pos) );
  }
  // remove all elements
  void removeAll(){  
    assert(this!=NULL);
    PetscFunctionBegin;
    // destroy elements
    CNode<T>* pNode;
    for (pNode = pNodeHead_; pNode != NULL; pNode = pNode->pNext)
      destructElements(&pNode->data, 1);

    nCount_ = 0;
    pNodeHead_ = pNodeTail_ = pNodeFree_ = NULL;
    pBlocks_->freeDataChain();
    pBlocks_ = NULL;
    PetscStackPop; /* since function returns void cannot use PetscFunctionReturn(); */
  }
  // iteration
  Position getHeadPosition() const{
    if(this==NULL){
      PetscTraceBackErrorHandler(1,"PromList::getHeadPosition","hh","prom_list.",2,3,"77",(void*)"8");
      return NULL; 
    }
    return (Position) pNodeHead_;
  } 
  Position  getTailPosition() const{
    assert(this!=NULL);
    return (Position) pNodeTail_;
  }
  T getNext(Position& rPosition) const{
    assert(this!=NULL);
    CNode<T>* pNode = (CNode<T>*) rPosition;
    assert(pNode!=NULL);
    rPosition = (Position) pNode->pNext; //assert(pNode->data!=(T)0);
    return pNode->data;
  } 
  T getPrev(Position& rPosition) const{
    assert(this!=NULL);
    CNode<T>* pNode = (CNode<T>*) rPosition;
    assert(pNode!=NULL);
    rPosition = (Position) pNode->pPrev; //assert( pNode->data != (T)0 );
    return pNode->data;
  }
  // getting/modifying an element at a given Position
  T getAt( Position pos ) const{
    assert(this!=NULL);
    CNode<T>* pNode = (CNode<T>*) pos;
    assert(pNode!=NULL); assert( pNode->data != (T)0 );
    return pNode->data;
  }
  void setAt( Position pos, T newElement ){
    assert(this!=NULL);
    CNode<T>* pNode = (CNode<T>*) pos;
    assert(pNode!=NULL); assert( pNode->data != (T)0 );
    pNode->data = newElement;
  }
  //
  T removeAt( Position pos ){
    assert(this != NULL);
    
    CNode<T>* pOldNode = (CNode<T>*) pos;
    assert(pOldNode!=NULL);
    T ret = pOldNode->data;
    
    // remove pOldNode from list
    if (pOldNode == pNodeHead_) {
      pNodeHead_ = pOldNode->pNext;
    }
    else {
      assert(pOldNode->pPrev!=NULL);
      pOldNode->pPrev->pNext = pOldNode->pNext;
    }
    if (pOldNode == pNodeTail_) {
      pNodeTail_ = pOldNode->pPrev;
    }
    else {
      assert(pOldNode->pNext!=NULL);
      pOldNode->pNext->pPrev = pOldNode->pPrev;
    }
    freeNode(pOldNode);
    return ret;
  }
  // inserting before or after a given Position
  Position insertBefore(Position pos, T newElement) {
    assert(this != NULL);  assert(newElement != 0);
    PetscFunctionBegin;
    if (pos == NULL)
      return addHead(newElement); // insert before nothing -> head of the list
    
    // Insert it before Position
    CNode<T>* pOldNode = (CNode<T>*) pos;
    CNode<T>* pNewNode = newNode(pOldNode->pPrev, pOldNode);
    pNewNode->data = newElement;
    
    if (pOldNode->pPrev != NULL){
      assert(pOldNode->pPrev!=NULL);
      pOldNode->pPrev->pNext = pNewNode;
    }
    else {
      assert(pOldNode == pNodeHead_);
      pNodeHead_ = pNewNode;
    }
    pOldNode->pPrev = pNewNode;
    PetscFunctionReturn((Position)pNewNode);
  }
  //
  Position insertAfter(Position pos, T newElement){
    assert(this != NULL);  assert(newElement != (T)0);
    PetscFunctionBegin;
    if (pos == NULL)
      return addTail(newElement); // insert after nothing -> tail of the list
    
    // Insert it before Position
    CNode<T>* pOldNode = (CNode<T>*) pos;
    assert(pOldNode!=NULL);
    CNode<T>* pNewNode = newNode(pOldNode, pOldNode->pNext);
    pNewNode->data = newElement;
    
    if (pOldNode->pNext != NULL){
      assert(pOldNode->pNext!=NULL);
      pOldNode->pNext->pPrev = pNewNode;
    }
    else {
      assert(pOldNode == pNodeTail_);
      pNodeTail_ = pNewNode;
    }
    pOldNode->pNext = pNewNode;
    PetscFunctionReturn((Position)pNewNode);
  }
  // helper functions (note: O(n)
  Position find(const T searchValue, Position startAfter = NULL) const{
    assert(this != NULL);
    
    CNode<T>* pNode = (CNode<T>*) startAfter;
    if (pNode == NULL) {
      pNode = pNodeHead_;  // start at head
    }
    else {
      pNode = pNode->pNext;  // start after the one specified
    }
    
    for (; pNode != NULL; pNode = pNode->pNext) {
      if( pNode->data  == searchValue ) return (Position)pNode;
    }
    return NULL;
  }
  // Implementation
  /////////////////////////////////////////////////////////////////////////////
  // Node helpers
  //
  // Implementation note: CNode's are stored in CPlex blocks and
  //  chained together. Free blocks are maintained in a singly linked list
  //  using the 'pNext' member of CNode with 'pNodeFree_' as the head.
  //  Used blocks are maintained in a doubly linked list using both 'pNext'
  //  and 'pPrev' as links and 'pNodeHead_' and 'pNodeTail_'
  //   as the head/tail.
  //
  // We never free a CPlex block unless the List is destroyed or removeAll()
  //  is used - so the total number of CPlex blocks may grow large depending
  //  on the maximum past size of the list.
  //
protected:  
  CNode<T>* newNode(CNode<T> *pPrev, CNode<T> *pNext ) {
    PetscFunctionBegin;
    if (pNodeFree_ == NULL) {
      // add another block
      CPlex* pNewBlock=CPlex::create(pBlocks_,nBlockSize_, sizeof(CNode<T>));
      // chain them into free list
      CNode<T>* pNode = (CNode<T>*) pNewBlock->data();
      // free in reverse order to make it easier to debug
      pNode += nBlockSize_ - 1;
      for (int i = nBlockSize_-1; i >= 0; i--, pNode--) {
	pNode->pNext = pNodeFree_;
	pNodeFree_ = pNode;
      }
    }
    assert(pNodeFree_ != NULL);  // we must have something
    CNode<T>* pNode = pNodeFree_;
    pNodeFree_ = pNodeFree_->pNext;
    pNode->pPrev = pPrev;
    pNode->pNext = pNext;
    nCount_++;
    assert(nCount_ > 0);  // make sure we don't overflow
    constructElements( &pNode->data, 1 );
    PetscFunctionReturn( pNode );
  }
  
  void freeNode( CNode<T> *pNode ) {
    destructElements( &pNode->data, 1 );
    pNode->pNext = pNodeFree_;
    pNodeFree_ = pNode;
    nCount_--;
    assert(nCount_ >= 0);  // make sure we don't underflow
  }

  void constructElements( T* pElements, int nCount ) {
    assert(pElements!=NULL);
    
    // default is bit-wise zero initialization
    memset((void*)pElements, 0, nCount * sizeof(T));
  }
  void destructElements(T* pElements, int nCount){
    assert(pElements!=NULL);
    // default does nothing
  }
  bool compareElements(T* pElement1, const T* pElement2) const {
    assert(pElement1!=NULL && pElement2!=NULL);
    return (*pElement1 == *pElement2);
  }
  //  data 
  CNode<T> *pNodeHead_;
  CNode<T> *pNodeTail_;
  CNode<T> *pNodeFree_;
  struct CPlex    *pBlocks_;
  int              nCount_;
  int              nBlockSize_;
};

#endif //__LIST_H__
