//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: prom_fine.C,v 1.8 2004/06/16 00:12:26 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           prom_fine.C
//

#include "prom_grid.hh"

/* FEAP quads - node numbers   - solids
 *                                __________-2-           
 *                            -3-/|        /|                          
 *                              / |       / |      
 *   -3- _________-2-      -7- /__|______/-6-       -3-
 *      |        |             |  |      |  |      |\---
 *      |  ___   |             |  |      |  |      |  \  ----_____-2- 
 *      |  |  |  |             |  |      |  |      |    \       / |
 *      | \|/ |  |             |  |      |  |      |      \  /    |
 *      |  `__|  |             |  |-0-   |  |      |      / \     |
 *      |        |             |  |______|__| -1-  |   /      \   |    
 *      |        |             | /       | /       |/           \ |
 *      |________|             |/________|/        ----------------
 *   -0-         -1-          -4-       -5-        -0-           -1- 
 *
 *         quad                   hex                  tet
 */ 

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::FineInit" 
/* PromGrid::FineInit **************************************************
 *
 *  - base grid constructor - setup from application's data sttructures
 *  - set fixity
 *
 *  INPUT: 
 *    - lid_ndof[numnd]: NULL if if constant block size 'bsize'
 *    - bsize:
 *    - gid_ghostLid:
 *    - elemList:
 * 
 *  SIDE EFFECTS:
 *    - form data structures of nodes and surfaces/solids
 *        * nodes lists - w/ adjac (of quads/solids)
 *        * sets nodes data structures
 *        * sets surface list
 *
 *  RETURN:
 */
int PromGrid::FineInit( const int lid_ndof[], int bsize, 
			const PromTable<int> &gid_ghostLid, 
			const PromList<int*> *elemList )
{
  int ierr,ii,jj,k,naf; 
  const int my0 = my0Nd(), nloc = nlocal(), next0 = my0 + nloc; 
  
  if(options_.verbose_>1){
    PetscPrintf((options_.verbose_>2) ? PETSC_COMM_SELF : gComm_,
		"[%d]%s:nNodes=%d,nLoc=%d,nGhostt=%d - nGlobal=%d -- Fine\n",
		comm_.mype(),__FUNCT__,nNodes_,nLocalNd_,nNodes_-nLocalNd_,
		nglobal());
  }
  
  // SET node fixity, get seed node (exterior node)
  ierr = CreateNodes( lid_ndof, bsize ); CHKERRQ(ierr); 
  // set adjacentcies  
  Position pos = elemList->getHeadPosition(); naf = 0;
  while( pos ) { 
    PromNode *pN[PROM_NENMAX], *curr, *n1;
    int lid, gid, *arr = elemList->getNext( pos );
    for( ii = 0 ; ii < PROM_NENMAX ; ii++ ){
      gid = arr[ii]; 
      if( gid < 0 ) break;             // sentila or low rank elem
      lid = (gid<my0 || gid>=next0) ? (gid_ghostLid.find(gid+1)-1) : (gid-my0);
      assert(lid>=0 && lid<nNodes_);
      pN[ii] = nodes_[lid];            // zero based
    }                                  // jjj has number of nodes
    const int nen = ii;
    //
    for( ii = 0 ; ii < nen ; ii++ ) { 
      curr = pN[ii];
      for ( jj = ii ; jj < nen ; jj++ ){  
	n1 = pN[jj]; 
	if( curr != n1 ) {
	  k = curr->symbData_->adjac.add( n1 );  if( k != 0 ) naf++;
	  k = n1->symbData_->adjac.add( curr );  if( k != 0 ) naf++;
	}
      }
    }
  }

  // out put
  if( options_.verbose_ > 1 ){
    MPI_Allreduce( &naf, &jj, 1, MPI_INT, MPI_SUM, gComm_ );
    PetscPrintf(gComm_, "[%d]%s %d nnz cells.\n",
		comm_.mype(), __FUNCT__, jj );
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::SetIndex" 
/* int PromGrid::SetIndex() ****************************************
 *
 *  - make maps between processor, global and FEAP ids, set node material.
 *
 *  INPUT:
 *    - 
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGrid::SetIndex( const PromTable<int> &gid_ghostLid ) 
{
  // set node indices < proc. pnode, gnode, fnode
  int xx, ierr, proc, gid, my0 = my0Nd();
  
  // set locals
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ) {
    PromNode *curr = nodes_[xx];
    PromLocalID *id = &curr->index_;
    id->setGlobal( my0 + xx );   // gid
  }
  // set ghost
  TablePos tpos = gid_ghostLid.getHeadPosition();  //assert(tpos!=0);
  while( tpos ) {
    xx = gid_ghostLid.getNext( tpos, &gid ); xx--; gid--;
    PromNode *curr = nodes_[xx]; assert(xx>=nLocalNd_ && xx<nNodes_);
    ierr = GetProc( gid, proc ); CHKERRQ(ierr); 
    curr->index_.setGlobal( gid );                     // gid
    int realid = gid - proc_gnode_[proc];
    curr->symbData_->ghostID.setLocal( realid );            // local id
    curr->symbData_->ghostID.setProc( proc );               // local id
    // add to table for fast lookup (key is one based)
    bool add = ghosts_.add( curr, gid + 1 );       assert( add );
  }
  assert( ghosts_.getCount() == nNodes_ - nLocalNd_ );
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGrid::CreateNodes" 
/* PromGrid::CreateNodes ******************************************
 *
 *  - Create nodes and fixity on nodes.
 *
 *  INPUT: 
 *    - lid_ndof[numnd]: NULL if if constant block size 'bsize'
 *    - bsize:
 *
 *  SIDE EFFECTS:
 *    - sets nodes, geq for locals (ghost never set?)
 *
 *  RETURN:
 *    - PETSc error code
 */
int PromGrid::CreateNodes( const int lid_ndof[], int bsize )
{
  int ierr,xx,ndf,geq; 

  for( xx = 0, geq = my0Eq() ; xx < nNodes_ ; xx++ ) {
    PromNode *curr = nodes_[xx];
    curr->index_ = xx;
    ndf = (lid_ndof==NULL) ? bsize : lid_ndof[xx]; assert(ndf>0);
    ierr = curr->SetNDF(ndf); CHKERRQ(ierr);
    if( xx < nLocalNd_ ){ curr->setGEQ(geq); geq += ndf; }
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::BuildElemList" 
/* PromGeoGrid::BuildElemList ******************************************
 *
 *  - Create elements from FE data stuctures
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 *    - Sets nodes adjacs.
 *
 *  RETURN:
 *    - PETSc error code
 */ 
int PromGeoGrid::BuildElemList( const PromTable<int> &gid_ghostLid, 
				const PromList<int*> *elemList )  
{
  // add elements to nodes elem list
  int ii,gid,lid,my0=my0Nd(),nloc=nlocal(),next0=my0+nloc;
  PromPoly *elem; PromNode    *pN[PROM_NENMAX],*curr;
  
  Position pos = elemList->getHeadPosition();
  while( pos ) { 
    int *arr = elemList->getNext( pos );     
    for( ii = 0 ; ii < PROM_NENMAX ; ii++ ) { 
      gid = arr[ii]; 
      if( gid < 0 ) break;    // low rank elem
      lid = (gid<my0 || gid>=next0) ? (gid_ghostLid.find(gid+1)-1) : (gid-my0);
      assert(lid>=0);
      assert( lid < nNodes_ );
      pN[ii] = nodes_[lid];              // zero based
    }                                      // jjj has number of nodes
    const int nen = ii, matid = -gid; // sentinal w/ material ID

    // add new elem. to nodes elem list -- GET RID OF DUPS 
    for( ii = 0 ; ii < nen ; ii++ ) {
      curr = pN[ii];      // set node types (element connected to)
      if( nen == 8 ) curr->setType( PROM_SOLID ); // 3D brick
      else if( nen == 4 && curr->getNDF() != 6 ){
	// hack to differentiate between tets and shell!!! -- membranes!!!
	if( curr->getType() == PROM_PLANE ) curr->setMaterialTopo();
	curr->setType( PROM_SOLID );
      }
      else if( nen == 4 ) {
	// 'solid' dominates 'plane' ???
	if( curr->getType() != PROM_SOLID ) curr->setType( PROM_PLANE ); 
	else curr->setMaterialTopo();
      }
      else curr->setType( PROM_N_POLY );
    }

    // create elemets grid elems (tets)
    if( nen == 8 ) { elem = new PromHex( pN ); CHKNEW(elem,1); }
    else if ( nen ==  4 ){ 
      if( pN[0]->getNDF() != 6 ) { 
	PromDTet *tet = new PromDTet(pN[0],pN[1],pN[2],pN[3]); CHKNEW(tet,1);
	elem = tet;	//assert( tet->getVolume() > 0.0 );
      }
      else {
	PromShell *shell = new PromShell( pN );  CHKNEW(shell,1);
	elem = shell;
      }
    }
    else {
      elem = new PromNPoly( pN, nen ); CHKNEW(elem,1); 
    }
    elem->setMaterialID( matid );    // feaps materail id
    elemList_.addTail( elem );     // add to list allways
    elem->AddSelfToNodes();  
  } // FEAP elements
  
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "PromGeoGrid::MakeElemAdjacs" 
/* PromGeoGrid::MakeElemAdjacs ******************************************
 *
 *  - sets (solid) element adjacentcies
 *
 *  INPUT: 
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::MakeElemAdjacs()
{ 
  PromTable<PromPoly*> table(117); 
  Position   posx, pos;
  PromPoly   *elem,*el,*elem2;
  int        j;
  PromNode *n1, *n2,*n3,*node1,*node2,*node3;
  
  // set elems adjac and make face list
  posx = elemList_.getHeadPosition(); //assert(posx!=NULL);
  while ( posx ) {
    elem = elemList_.getNext(posx);
    
    if ( elem->getType() == PROM_SOLID ){
      if( elem->numExposedFaces() == 0 ) continue;
    }
    else continue;
  
    // get all adjacent elems
    assert( table.isEmpty() ); 
    for ( j = 0 ; j < elem->numNodes() ; j++ ) {
      n1 = dynamic_cast<PromNode*>(elem->nodes_[j]);
      pos = n1->symbData_->elemList.getHeadPosition();
      while ( pos ) {
	el = n1->symbData_->elemList.getNext(pos);
	if ( el != elem ) table.add(el);
      }
    }
    
    // set 'ajac' for elem
    const int numadj = elem->numNeigh(); 
    const int *elID = elem->getIDArr(); if(!elID)continue;
    while ( !table.isEmpty() ) {
      elem2 = table.removeHead();
      if( elem2->getType() != elem->getType() || 
	  elem2->numNodes() != elem->numNodes() ) continue; 
      const int numadj2 = elem2->numNeigh();
      const int *elID2 = elem2->getIDArr(); assert(elID2!=NULL);
      
      // see if elem2 fits on each face of elem
      int found = 0;
      for ( int i = 0 ; i < numadj && !found ; i++ ) {
	if ( elem->neigh_[i] ) continue;
	int id = elID[ 4 * i + 3 ];
	const int N1 = id == -9 ? 3 : 4 ;
	node1 = elem->nodes_[ elID[ 4 * i + 2 ] ];
	node2 = elem->nodes_[ elID[ 4 * i + 1 ] ];
	node3 = elem->nodes_[ elID[ 4 * i  ] ];
	
	// try each face of test elem          
	for ( j = 0 ; j < numadj2 && !found ; j++ ) {
	  if ( elem2->neigh_[j] ) continue;
	  const int N2 = elID2[ 4 * j + 3 ] == -9 ? 3 : 4 ;
	  if ( N1 != N2 ) continue;
	  int indexj = 4 * j;
	  for ( int jj = 0 ; jj < N2 && !found ; jj++ ) {
	    // nodes to look for on test face 
	    n1 = elem2->nodes_[ elID2[ indexj + (jj)%N2 ] ];
	    n2 = elem2->nodes_[ elID2[ indexj + (jj+1)%N2 ] ];
	    n3 = elem2->nodes_[ elID2[ indexj + (jj+2)%N2 ] ];
	    if ( n1 == node1 && n2 == node2 && n3 == node3 ) {
	      found = 1;
	      elem->neigh_[i] = elem2;
	      elem2->neigh_[j] = elem;
	    }
	  }
	}
      }
    }
  }

  return 0;
}

